'\" 
'\" $Header: /home/neil/src/tcl/tensor/doc/RCS/tensor.n,v 1.5 2008/07/19 20:30:33 neil Exp $
'\" 
.so man.macros
.HS tensor tcl
.BS
'\" Note:  do not modify the .SH NAME line immediately below!
.SH NAME
tensor \- Create a tensor object
.SH SYNOPSIS
.sp
\fBpackage require Tensor\fR
.sp
\fBtensor::create \fI tensorName \fR?\fB\-type \fItensorType\fR? \fR?\fB\-size \fIsize\fR? \fR?\fB\-initial \fIinitList\fR? \fR?\fB\-initfile \fIfileName\fR? \fR?\fB\-initchannel \fIfileChannel\fR?
.sp
\fBtensor::exists \fI tensorName\fR
.BE

.SH DESCRIPTION
.PP
The command \fB tensor::create \fI tensorName \fR creates a
tensor object named \fItensorName\fR. \fBtensor::exists \fItensorName\fR
returns a Boolean value indicating whether or not a tensor
called \fItensorName\fR exists.
.PP
Tensor objects may be thought of as multidimensional, rectangular
arrays of numerical data. Legal tensor
types are \fBfloat\fR, \fBdouble\fR, \fBint\fR, \fBshort\fR, \fBbyte\fR,
and \fBunsigned\fR,
where types \fBshort\fR, \fBbyte\fR, and (of course) \fBunsigned\fR
are unsigned. If the \fB\-type\fR
option is omitted, it defaults to \fBdouble\fR. The command must have exactly
one of the \fB\-size\fR, \fB\-initial\fR, or \fB\-initfile\fR options. If
\fB\-size\fR is used, it must be followed by a list of 0 to 16 positive integers,
giving the dimensions of the tensor; the length of the size list is the
\fIorder\fR of the tensor. For example, the command
.DS
\fBtensor::create x345 -size {3 4 5}\fR
.DE
declares a tensor of type double (the default) whose order is 3 and whose
dimensions are 3x4x5. The \fB\-initial\fR option must be followed by a
legal Tcl list containing initialization data for the tensor. The order of
the tensor is determined from the nesting level of the list, and the lengths
of the list and (if the order is greater than 1) all the list's sublists
determine the size of the tensor. The command
.DS
\fBtensor::create I33 -type float -initial { { 1 0 0 } { 0 1 0 } { 0 0 1 } }\fR
.DE
declares I33 to be a 3x3 float tensor which is initialized to the
3x3 identity matrix. The \fB\-initfile\fR and \fB-initchannel\fR
options must be followed by a file name or a Tcl file handle
(as returned by \fBopen\fR), respectively. In these cases, the tensor's type,
order, size, and contents are read from the given file.

.SH TENSOR OBJECT COMMANDS

.PP
Every tensor declaration creates an associated command, the tensor object
command. Thus the examples above create commands named ``x345'' and ``I33''.
The functions performed by the tensor object command can be divided into five
broad categories: inquiry functions, file I/O functions, storage rearrangement
functions, arithmetic operations, and comparison operations.
The inquiry functions are:
.TP
\fItensorName \fBtype\fR
Returns the type of the tensor, i.e. \fBdouble\fR, \fBfloat\fR, \fBint\fR,
\fBshort\fR, \fBbyte\fR or \fBunsigned\fR.
.TP
\fItensorName \fBorder\fR
Returns the order of the tensor, i.e. the number of dimensions or indices
.TP
\fItensorName \fBdimensions\fR
Returns the dimensions of the tensor, formatted as a Tcl list.
.TP
\fItensorName \fR?\fBsection\fI section_spec\fR?
Returns the contents of the tensor, or some section of it, as a nested
Tcl list. The \fIsection_spec\fR is a list, whose length must equal the order
of the tensor. Each element of the list specifies a range of index values,
and must be either a list containing two non-negative integers (e.g. ``{2 5}''
to indicate that the index runs from 2 through 5 inclusive), a single number,
or ``*''. A single number N is equivalent to the list {N N}, and ``*'' is the
same as the range ``{0 M-1}'' where M is the index range for the corresponding
dimension. (Note that indexing is zero-based, as it is in C and in Tcl lists.)
.TP
\fItensorName \fR?\fBsection\fI section_spec\fR?
Returns the contents of the tensor, or some section of it, in storage
order.
.PP
There are two I/O commands. They are:
.TP
\fItensorName \fR?\fBsection\fI sectionSpec\fR? \fBwrite\fR \fIfile\fR ?\fIfileFormat\fR?
Writes the tensor (or section of the tensor) out to a file. 
\fIfileFormat\fR may be \fBbinary\fR (which is the default), or it may be
\fBascii\fR or \fBtext\fR, which are synonymous, and force writing of the tensor
file in a text format. (Note that the binary tensor file format is NOT machine
independent; if you plan on sending a tensor file to a machine with a different
architecture, you should send a text-format file.)
.TP
\fItensorName \fR?\fBsection\fI sectionSpec\fR? \fBwritechannel\fR \fIchannel\fR ?\fIfileFormat\fR?
This is exactly like the \fBwrite\fR command, except that it writes to an
already-open Tcl channel rather than a file.

.PP
Although it's not really an ``I/O'' command, the \fBput\fR command does
move data from the tensor to another place, namely a Tcl associative array:
.TP
\fItensorName \fR?\fBsection\fI sectionSpec\fR? \fBput into\fI arrayName\fR ?\fBif\fI compOp number \fR? ?\fB-offset\fI offset\fR?
copies the data in the specified section of the tensor into array
\fIarrayName\fR. By default, the indices into the array will be the same as the
tensor indices, i.e. if we put section {{2 3} {3 5}} of an order-2 tensor
into an array, then the array indices will be (2,3), (2,4), (2,5),
(3,3,), (3,4), and (3,5). (Note that the array indices are comma-separated.)
The \fBif\fR clause allows only selected elements to be copied to the array.
\fIcompOp\fR may be any of the normal C comparison operations; the array element
will only be written if the comparison holds true.
The \fB-offset\fR option allows this behavior to be changed. For example,
specifying an offset of \fB{1 1}\fR for an order-two tensor will put tensor
element {0 0} into Tcl array element (1,1).
.PP
There are seven storage rearrangement functions:
.TP
\fItensorName \fBflipindices \fIindexList\fR
``flips'' the indices listed in \fIindexList\fR, i.e. reverses the order
of the data for those indices. For example, if an order-1 tensor named \fBz\fR
with three
elements contains the data { 1 2 3 }, then after performing the command
\fBz flipindices 0\fR it will
contain { 3 2 1 }.
.TP
\fItensorName \fBpermuteindices \fIpermuteList\fR
permutes the indices of the tensor. The length of \fIpermuteList\fR
must
be equal to the order N of the tensor, and it must contain each of the
integers 0 through N-1 exactly once. The number found in position \fIi\fR
of the list (call it \fIj\fR) indicates that index \fIi\fR after the
\fBpermuteindices\fR operation will correspond to index \fIj\fR before
the operation. For instance, after performing
\fItensorName \fBpermuteindices { 0 2 1 }\fR
on an order-3 tensor, data that
used to be found at location (i,j,k) will now be found at location (i,k,j).
.TP
\fItensorName \fBreshape \fIdimensionList\fR
changes the dimensions of \fItensorName\fR to the dimensions
given by \fIdimensionList\fR. \fIdimensionList\fR is a Tcl list
of positive integers. The new dimensions must not change the number
of elements in the tensor; for example, if \fBx\fR is a tensor of order
2, size 20 by 3, we may say \fBx resize 60\fR
and \fBx resize {5 3 4}\fR,
but we may not say \fBx resize {10 3}\fR,
since this would produce a tensor with 30 elements rather that 60.
Reshaping a tensor maintains the storage order of the elements.
.TP
\fItensorName \fBswap \fIdim swapList\fR
swaps cross-sections of a tensor. \fIdim\fR specifies the dimension
along which we're swapping: 0 for rows, 1 for columns, and so forth.
\fIswapList\fR is a two-element list containing the indices of the
two swapped elements. To enhance readability,
\fIdim\fR may also be the words \fBrow\fR or \fBrows\fR (equivalent to 0),
or the words\fBcol\fR, \fBcols\fR, \fBcolumn\fR, or \fBcolumns\fR
(equivalent to 1).
.TP
\fItensorName \fBdelete \fIdim deleteList\fR
deletes elements along the selected dimension. \fIdim\fR takes the same
form as the \fIdim\fR argument of the \fBswap\fR command; \fIdeleteList\fR
is a list of elements to delete.
.TP
\fItensorName \fBinsert \fInum dim \fR{\fBbefore\fR|\fBafter\fR} \fIpos\fR
inserts elements along the selected dimension. \fInum\fR is the number
of elements to insert; \fIdim\fR takes the same form as the \fIdim\fR
argument of the \fBswap\fR command. \fIpos\fR is the position where we'll
insert the rows/columns. The inserted data will be equal to zero, and will
be inserted before or after \fIpos\fR, depending on the preceding keyword.
\fIpos\fR may be either an integer or the word \fBend\fR, denoting the last
row/column in the tensor. The contents of the tensor will ``slide down''
to make room for the inserted data.
.TP
\fItensorName \fBpermute \fIdim permuteList\fR
permutes elements along the selected dimension. \fIdim\fR takes the same
form as the \fIdim\fR argument of the \fBswap\fR command; \fIpermuteList\fR
is a list that specifies how the tensor's rows/columns will be permuted.
\fIpermuteList\fR must contain the numbers 0 through \fIN\fR-1, where \fIN\fR
is the extent of the chosen dimension. At position \fIi\fR, the permuted
tensor will contain the data previously at position \fIP(i)\fR, where
\fIP(i)\fR is the \fIi\fRth element of \fIpermuteList\fR.
\fIpermuteList\fR may be either a Tcl list or a tensor of order 1.

.PP
There are a variety of arithmetic operations on tensors, including scalar
and tensor assignment, addition, subtraction, multiplication, and division.
(See the \fBtensorexpr\fR manual page for an alternative way to perform
many arithmetic operations via the \fBtensor::expr\fR command.)
The syntax for the most commonly-used arithmetic object commands is
.RS
  \fIdestTensor\fR ?\fBsection\fI sectionSpec\fR? \fIop= operand\fR
.RE
where \fIop=\fR may be =, +=, -=, *=, or /=, which are analogous to the same
operators in C. (One important difference from C is that attempts to divide
by zero are silently ignored.)
\fIoperand\fR may have a variety of forms. The most commonly used ones are
.RS
  \fBscalar \fInumber\fR
.RE
which performs the specified operation between each of the specified elements
of \fIdestTensor\fR and the given number, and
.RS
  \fBtensor \fIsrcTensor\fR ?\fBsection\fI section_spec\fR?
.RE
which performs the specified operation between each of the specified elements
of \fIdestTensor\fR and the corresponding element of \fIsrcTensor\fR. Note that
the specified sections of the source and destination tensors must have the
same order and dimensions.
.PP
Other operand directives behave like the \fBtensor\fR operand type; they
produce a multidimensional array, and the operation is performed on
corresponding elements of the destination and source arrays. The other
allowable operands are:
.TP
\fBarray \fIdata\fR
where \fIdata\fR is a (possibly nested) Tcl list
containing numerical data. The format is the same as for the \fB\-initial\fR
option of the \fBtensor\fR command.
.TP
\fBfile \fIfileSpec\fR
where \fIfileSpec\fR has the same format as the
\fB\-initfile\fR option of the \fBtensor\fR command.
.TP
\fBcontraction \fItensorName indexList\fR
where \fItensorName\fR is the
tensor whose contraction is to be computed, and \fIindexList\fR is the
list of indices over which summation is performed. If \fItensorName\fR
is of order M and \fIindexList\fR is of length N, then the contraction
operation yields a tensor whose order is M-N and whose dimensions are
given by the the indices of \fItensorName\fR which remain when the indices
given in \fIindexList\fR are dropped. All indices in \fIindexList\fR
must have the same dimension.
.TP
\fBmultensor \fItensorName1 tensorName2 indexPair\fR
where \fItensorName1\fR
and \fItensorName2\fR are the tensors to be multiplied, and \fIindexPair\fR
is a list containing the two indices (in the first and second tensors)
with respect to which multiplication and summation take place.
.TP
\fBouter \fItensorName1 tensorName2\fR
where \fItensorName1\fR
and \fItensorName2\fR are the tensors whose outer product is computed.
.TP
\fBboolean \fItensorName1\fR ?\fBsection\fI sectionSpec\fR? \fIcompOp operand\fR
where \fItensorName1\fR is the name of a tensor, \fIcompOp\fR is a comparison
operator, and \fIoperand\fR is either \fBscalar\fI number\fR or
\fBtensor \fItensorName2\fR ?\fBsection\fI sectionSpec\fR?.
This operand will perform an elementwise Boolean operation on
\fItensorName1\fR, and create a tensor of the appropriate size containing
only zeroes and ones.
.PP
It is also possible to apply a function to all or part of a tensor, via
the \fBapply\fR tensor object command. Its syntax is:
.RS
  \fIdestTensor\fR ?\fBsection\fI sectionSpec\fR? \fBapply\fI functionName\fR
.RE
where \fIfunctionName\fR may be \fBsin\fR, \fBcos\fR, \fBtan\fR,
\fBasin\fR, \fBacos\fR, \fBatan\fR,
\fBsinh\fR, \fBcosh\fR, \fBtanh\fR,
\fBasinh\fR, \fBacosh\fR, \fBatanh\fR,
\fBsqrt\fR, \fBexp\fR, \fBlog\fR, \fBlog10\fR,
\fBfloor\fR, \fBceil\fR, or \fBfabs\fR.
All elements in the specified tensor section (or the whole tensor, if no section
was given) will have the given function applied to them, and the original value
will be replaced with the function result.
.PP
There are three other tensor arithmetic operations: \fBsaturate\fR,
\fBsequnce\fR,
and \fBsubstitute\fR. The \fBsaturate\fR operation has the syntax
.RS
  \fIdestTensor\fR ?\fBsection\fI sectionSpec\fR? \fBsaturate\fR \fIlowValue highValue\fR
.RE
This operation ``saturates'' all elements of the specified tensor so that
their values lie between \fIlowValue\fR and \fIhighValue\fR, i.e. any element
that is less than \fIlowValue\fR will be replaced with \fIlowValue\fR, and
any element greater than \fIhighValue\fR will be replaced with \fIhighValue\fR.
The \fBsequence\fR command sets the elements of a section of a tensor
to sequentially increasing values, starting at 0.
The syntax of the \fBsequence\fR command is
.RS
  \fIdestTensor\fR ?\fBsection\fI sectionSpec\fR? \fBsequence\fR
.RE
For instance, if \fBt55\fR is a 5-by-5 tensor and is initially set to zero,
then the command
.RS
  \fBt55 section {{1 3} {1 3}} sequence\fR
.RE
will place the numbers zero through eight into the ``middle'' of the tensor,
i.e. the tensor's contents will be
.RS
  { 0 0 0 0 0 }
  { 0 0 1 2 0 }
  { 0 3 4 5 0 }
  { 0 6 7 8 0 }
  { 0 0 0 0 0 }
.RE
The \fBsubstitute\fR command takes the form
.RS
  \fIdestTensor\fR ?\fBsection\fI sectionSpec\fR? \fBsubstitute\fR \fIsubstTable\fR
.RE
where \fIsubstTable\fR may be either the name of an order-1 tensor, or a Tcl
list of floating-point numbers. For each of the specified elements of
\fIdestTensor\fR, a value from \fIsubstTable\fR will be substituted. The value
is chosen by truncating the element's value to an integer, and indexing into
\fIsubstTable\fR. If the element is negative, the first element of
\fIsubstTable\fR will be used; if it is greater than the length of the table,
then the last element will be used.
.PP
The comparison operations are implemented in the same way as, and take the
same operands as, the arithmetic assignment operations. They have the same syntax,
except
that the legal operators are ==, !=, <, <=, >, and >=. But unlike the arithmetic
operations, they don't affect the values stored in the tensor; instead, they
return a value of either 1 or 0, depending upon whether the comparison
operation was or was not true for ALL corresponding elements in the tensor
and its argument. Note that the rules about logical negation that work for
scalar comparisons don't work for tensor comparisons; for scalars,
!(xyz == 0) is equivalent to (xyz != 0); this is not the case for tensor
comparisons. If we have a tensor xyz of order 1 with three
elements, say { 0 0 1 }, then
.DS
  xyz == scalar 0
.DE
will return 0, since the comparison fails for the third element; likewise,
.DS
  xyz != scalar 0
.DE
also returns 0, since the comparison fails for the first and second elements.
.PP
The \fBlistindices\fR command is essentially a variation on the comparison
operators. Its syntax is
.RS
  \fIdestTensor\fR ?\fBsection\fI sectionSpec\fR? \fBlistindices\fR ?\fIcomponentList\fR? \fBwhere\fR \fIcomparisonOp operand\fR
.RE
where \fIcomparisonOp\fR may be any valid comparison operator, and \fIoperand\fR
is the same as for the comparison operators. This command returns a list of
tensor section indices where the given comparison is true. \fIcomponentList\fR is
a list of integers indicating which components of the section index should be
reported. This is useful if, for example, you are only interested in the
column indices for a given row of the tensor. The default value of \fIcomponentList\fR is
the list \fB{0 1 2 ... \fIN-1\fB}\fR, where \fIN\fR is the order of the tensor.
The components are reported in the order given by fIcomponentList\fR, so it is
also possible to permute and/or duplicate the components.
.PP
Although not strictly a ``comparison operation'', tensor objects also
have a \fBminmax\fR command:
.DS
  \fItensorName \fR ?\fBsection\fI section_spec\fR? \fBminmax\fR ?\fB-indices\fR?
.DE
This command returns a list containing two
elements: the minimum and maximum values contained in the specified
section of the tensor. If the \fB-indices\fR switch is given, then
instead of the minimum and maximum values, it returns the tensor indices
where the minimum and maximum values occur.
.SH EXAMPLES

.PP
Typical matrix operations are easily performed with tensor objects. For
example, to multiply a 5-by-3 matrix \fBA\fR by a 3-by-2 matrix \fBB\fR,
we could do the following:
.RS
    tensor A -size {5 3}
    tensor B -size {3 2}
    A = array { {1 5 3} {4 8 2} {3 7 6} {1 4 3} {5 5 5} }
    B = array { {4 2} {2 4} {9 1} }

    tensor C -size {5 2}
    C = multensor A B {1 0}
.RE
The \fB{1 0}\fR index pair indicates that index 1 of tensor A and
index 0 of tensor B will be indexed over during the multiplication.
(Note that if B were a 2-by-3 matrix instead of 3-by-2, we could have
used \fB{1 1}\fR as the index pair, which would be equivalent of
multiplying A by the \fItranspose\fR of B.)
.PP
Summing along the rows or columns of a matrix also is easy; we use the
"contraction" operation. To sum the rows of a matrix \fBA\fR to form a
vector \fBS\fR, we could do the following (if A is declared as above):
.RS
    tensor S -size 5
    S = contraction A 1
.RE
This does a running sum, varying the \fIcolumn\fR index,
thus summing along each \fIrow\fR. A matrix trace is also easy:
.RS
    tensor M33 -initial { { 1 2 5 } { 3 1 4 } { 9 0 -1 } }
    tensor T -size {}
    T = contraction M33 {0 1}
.RE
This declares T to be a \fIscalar\fR, since it has no dimensions,
and sets the contents of T by adding up M33(0,0) + M33(1,1) + M33(2,2).
Note that we sum along \fIboth\fR indices 0 and 1.
.PP
Many operations work on sections of tensors, as well as entire
tensors. We may, for instance, set only a section of a tensor in
an assignement operation:
.RS
    tensor D -size {5 7}
    tensor X -size 7
    D section {1 *} = scalar 2
    D section {{2 4} {1 4}} = scalar 3
    X section {{1 3}} = scalar -6
.RE
These operations set row 1 of D (the \fIsecond\fR row, because of zero-based
array indexing) to 2, and sets the submatrix consisting of columns 1 through 4,
rows 2 through 4, to 3. Finally, we set elements 1 through 3 of X to -6.
(Note that we need an extra pair if braces here. Since X is 1-dimensional,
the section is described by a list of length \fIone\fR, but the single
element of this list is of length \fItwo\fR, since we're specifying the
upper and lower bounds of a range.)
.PP
The storage rearrangement functions are fairly straightforward. For example,
.RS
    t swap 0 {3 5}
.RE
will exchange rows 3 and 5 of tensor \fBt\fR.
.RS
    t delete 1 {2 4 7}
.RE
will delete columns 2, 4, and 7 of tensor \fBt\fR. The
.RS
    t insert 2 columns after 1
.RE
will insert 2 columns into tensor \fBt\fR; the new columns will be columns
2 and 3, while the old column 2 will become column 4, the old column 3
will become column 5, and so forth.
The command
.RS
    t permute columns {0 1 2 5 3 4}
.RE
will rearrange the columns of tensor \fBt\fR so that columns 0, 1, and 2
remain in position, the new column 3 is the old column 5, the new column
4 is the old column 3, and the new column 5 is the old column 4.
.PP
The \fBput into\fR tensor command can be used to put the contents of a tensor
into an array; this is fairly straightforward. However, sometimes we want
to selectively write into an array. This can be done with the \fBif\fR clause.
For instance, if we want to write only those elements of tensor \fBt\fR
that are positive, we could say
.RS
    t put into a if > 0
.RE
.PP
A typical function application is:
.RS
    t section {0 *} apply fabs
.RE
This will replace the first row of tensor ``t'' with its absolute value.

.SH AUTHOR
Neil McKay, mckay@eecs.umich.edu

.SH "SEE ALSO"
tensorexpr(n), tensor(3)

.SH COPYRIGHT

.PP
Original document Copyright 2002 by General Motors Corporation.
All rights reserved.
.PP
Modifications Copyright 2008 by Neil McKay, who disclaims any liability
associated with the use of this software, but places
no additional restrictions on distribution of this file.
.PP
The following terms apply to all files associated with the software
unless otherwise explicitly stated in the individual files.
.PP
Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided
that the above copyright notice appears in all copies, and that both
the copyright notice and warranty disclaimer appear verbatim in all
copies and supporting documentation, and that the name and trademarks
of General Motors Corporation or any of its entities not be used in
advertising or publicity pertaining to the software without specific,
written prior permission.
.PP
No written agreement, license or royalty fee is required for any of the
authorized uses.  Modifications to the software may be copyrighted by
their authors and need not follow these licensing terms, provided the
new terms are clearly indicated on the first page of each file to which
they apply.
.PP
GENERAL MOTORS CORPORATION DISCLAIMS ANY AND ALL EXPRESS OR IMPLIED
WARRANTIES AND CONDITIONS OF ANY KIND WITH REGARD TO THIS SOFTWARE,
INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY WARRANTIES OR CONDITIONS
OF TITLE OR NON-INFRINGEMENT.  GENERAL MOTORS CORPORATION MAKES NO
REPRESENTATION THAT THE USE OF THE SOFTWARE OR DOCUMENTATION WILL NOT
INFRINGE ANY THIRD PARTY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS.
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, AND GENERAL MOTORS HAS
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
OR MODIFICATIONS OF ANY KIND.
.PP
IN NO EVENT SHALL GENERAL MOTORS CORPORATION BE LIABLE TO ANY PARTY FOR
ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, CONSEQUENTIAL,
OR OTHER DAMAGES WHATSOEVER ARISING OUT OF, OR IN CONNECTION WITH, THE
DISTRIBUTION, USE AND PERFORMANCE OF THIS SOFTWARE, ITS DOCUMENTATION, OR
ANY DERIVATIVES THEREOF, HOWEVER CAUSED, AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY OR TORT, EVEN IF GENERAL MOTORS
CORPORATION HAD BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
.PP
Each recipient, user and distributee is solely responsible for determining
the appropriateness of using, copying, modifying and distributing this
software and assumes all risks associated with the exercise of the rights
granted herein.

